(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('points.js')) :
	typeof define === 'function' && define.amd ? define(['points.js'], factory) :
	(global.morph = factory(global.points_js));
}(this, (function (points_js) { 'use strict';

function draw (m, p, ctx) {
	return function (t) {
	var v = m(t);
	ctx.clearRect(0,0,p.width,p.height);
	ctx.beginPath();
	for ( var item = (void 0), i = 0, len = v.length; i < len; i++) {
		item = v[i];
		if (item.moveTo) {
			ctx.moveTo(item.x, item.y);
		} else if (item.curve && item.curve.type === 'cubic') {
			ctx.bezierCurveTo(item.curve.x1, item.curve.y1, item.curve.x2, item.curve.y2, item.x, item.y);
		} else if (item.curve && item.curve.type === 'quadratic') {
			ctx.quadraticCurveTo(item.curve.x1, item.curve.y1, item.x, item.y);
		} else {
			ctx.lineTo(item.x, item.y);
		}
	}
	ctx.strokeStyle = p.fill || '#555';
	ctx.stroke();
	}
}

/** global: requestAnimationFrame */
/** global: setTimeout */
var tasks = [];
var DONE = -1;
var raf = typeof requestAnimationFrame !== "undefined" ? requestAnimationFrame : typeof setTimeout !== "undefined" ? function (fn) { return setTimeout(fn, 16); } : function () {};
raf(function _raf(time) {
	raf(_raf);
	if (tasks.length === 0) {
		return false;
	}
	var i = 0;
	while (i < tasks.length) {
		if (tasks[i](time) === DONE) {
			tasks.splice(i, 1);
		} else {
			i++;
		}
	}
	return true;
});

function frame (ref) {
	var frame = ref.frame;
	var duration = ref.duration; if ( duration === void 0 ) duration = 1000;
	var startTime = ref.startTime; if ( startTime === void 0 ) startTime = 0;
	var repeat = ref.repeat; if ( repeat === void 0 ) repeat = 0;
	var delay = ref.delay; if ( delay === void 0 ) delay = 0;
	var repeatDelay = ref.repeatDelay; if ( repeatDelay === void 0 ) repeatDelay = 0;
	var yoyo = ref.yoyo; if ( yoyo === void 0 ) yoyo = false;
	var reversed = ref.reversed; if ( reversed === void 0 ) reversed = false;
	var easing = ref.easing;


	startTime += delay + performance.now();

	tasks.push(function (time) {	

		if (time < startTime) {

			return false;

		}

		var elapsed = Math.min((time - startTime) / duration, 1);
		var value = easing ? easing(elapsed) : elapsed;

		if (reversed) {
			value = 1 - value;
			elapsed = 1 - elapsed;
		}

		frame(value, elapsed);

		if (elapsed === 1 || (reversed && elapsed === 0)) {

			if (repeat) {

				if (isFinite(repeat)) {
					repeat--;
				}

				if (yoyo) {

					reversed = !reversed;

				}

				startTime = time + repeatDelay;

			return true;

			} else {

			return DONE;

			}

		}

		return true;

	});
}

var middlewares = {};

middlewares.color = function (a) {
	return ("rgb(" + (a.map(function (v) { return v | 0; }).join(",")) + ")");
};

function middleware (a, b, r) {
	var _a = [].concat( a ), len = _a.length, gmw = typeof r === "string" && middlewares[r];
	return function (t) {
		for ( var i = 0; i < len; i++ ) {
			_a[i] = a[i] + (b[i] - a[i]) * t;
		}
		return gmw ? gmw(_a) : _a;
	}
}

var _extends = Object.assign || function (target) {
var arguments$1 = arguments;
 for (var i = 1; i < arguments.length; i++) { var source = arguments$1[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) { continue; } if (!Object.prototype.hasOwnProperty.call(obj, i)) { continue; } target[i] = obj[i]; } return target; }

var toPoints = function toPoints(_ref) {
  var type = _ref.type,
      props = _objectWithoutProperties(_ref, ['type']);

  switch (type) {
    case 'circle':
      return getPointsFromCircle(props);
    case 'ellipse':
      return getPointsFromEllipse(props);
    case 'line':
      return getPointsFromLine(props);
    case 'path':
      return getPointsFromPath(props);
    case 'polygon':
      return getPointsFromPolygon(props);
    case 'polyline':
      return getPointsFromPolyline(props);
    case 'rect':
      return getPointsFromRect(props);
    case 'g':
      return getPointsFromG(props);
    default:
      throw new Error('Not a valid shape type');
  }
};

var getPointsFromCircle = function getPointsFromCircle(_ref2) {
  var cx = _ref2.cx,
      cy = _ref2.cy,
      r = _ref2.r;

  return [{ x: cx, y: cy - r, moveTo: true }, { x: cx, y: cy + r, curve: { type: 'arc', rx: r, ry: r, sweepFlag: 1 } }, { x: cx, y: cy - r, curve: { type: 'arc', rx: r, ry: r, sweepFlag: 1 } }];
};

var getPointsFromEllipse = function getPointsFromEllipse(_ref3) {
  var cx = _ref3.cx,
      cy = _ref3.cy,
      rx = _ref3.rx,
      ry = _ref3.ry;

  return [{ x: cx, y: cy - ry, moveTo: true }, { x: cx, y: cy + ry, curve: { type: 'arc', rx: rx, ry: ry, sweepFlag: 1 } }, { x: cx, y: cy - ry, curve: { type: 'arc', rx: rx, ry: ry, sweepFlag: 1 } }];
};

var getPointsFromLine = function getPointsFromLine(_ref4) {
  var x1 = _ref4.x1,
      x2 = _ref4.x2,
      y1 = _ref4.y1,
      y2 = _ref4.y2;

  return [{ x: x1, y: y1, moveTo: true }, { x: x2, y: y2 }];
};

var validCommands = /[MmLlHhVvCcSsQqTtAaZz]/g;

var commandLengths = {
  A: 7,
  C: 6,
  H: 1,
  L: 2,
  M: 2,
  Q: 4,
  S: 4,
  T: 2,
  V: 1,
  Z: 0
};

var relativeCommands = ['a', 'c', 'h', 'l', 'm', 'q', 's', 't', 'v'];

var isRelative = function isRelative(command) {
  return relativeCommands.indexOf(command) !== -1;
};

var optionalArcKeys = ['xAxisRotation', 'largeArcFlag', 'sweepFlag'];

var getCommands = function getCommands(d) {
  return d.match(validCommands);
};

var getParams = function getParams(d) {
  return d.split(validCommands).map(function (v) {
    return v.replace(/[0-9]+-/g, function (m) {
      return m.slice(0, -1) + ' -';
    });
  }).map(function (v) {
    return v.replace(/\.[0-9]+/g, function (m) {
      return m + ' ';
    });
  }).map(function (v) {
    return v.trim();
  }).filter(function (v) {
    return v.length > 0;
  }).map(function (v) {
    return v.split(/[ ,]+/).map(parseFloat).filter(function (n) {
      return !isNaN(n);
    });
  });
};

var getPointsFromPath = function getPointsFromPath(_ref5) {
  var d = _ref5.d;

  var commands = getCommands(d);
  var params = getParams(d);

  var points = [];

  var moveTo = void 0;

  for (var i = 0, l = commands.length; i < l; i++) {
    var command = commands[i];
    var upperCaseCommand = command.toUpperCase();
    var commandLength = commandLengths[upperCaseCommand];
    var relative = isRelative(command);
    var prevPoint = i === 0 ? null : points[points.length - 1];

    if (commandLength > 0) {
      var commandParams = params.shift();
      var iterations = commandParams.length / commandLength;

      for (var j = 0; j < iterations; j++) {
        switch (upperCaseCommand) {
          case 'M':
            var x = (relative && prevPoint ? prevPoint.x : 0) + commandParams.shift();
            var y = (relative && prevPoint ? prevPoint.y : 0) + commandParams.shift();

            moveTo = { x: x, y: y };

            points.push({ x: x, y: y, moveTo: true });

            break;

          case 'L':
            points.push({
              x: (relative ? prevPoint.x : 0) + commandParams.shift(),
              y: (relative ? prevPoint.y : 0) + commandParams.shift()
            });

            break;

          case 'H':
            points.push({
              x: (relative ? prevPoint.x : 0) + commandParams.shift(),
              y: prevPoint.y
            });

            break;

          case 'V':
            points.push({
              x: prevPoint.x,
              y: (relative ? prevPoint.y : 0) + commandParams.shift()
            });

            break;

          case 'A':
            points.push({
              curve: {
                type: 'arc',
                rx: commandParams.shift(),
                ry: commandParams.shift(),
                xAxisRotation: commandParams.shift(),
                largeArcFlag: commandParams.shift(),
                sweepFlag: commandParams.shift()
              },
              x: (relative ? prevPoint.x : 0) + commandParams.shift(),
              y: (relative ? prevPoint.y : 0) + commandParams.shift()
            });

            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
              for (var _iterator = optionalArcKeys[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var k = _step.value;

                if (points[points.length - 1]['curve'][k] === 0) {
                  delete points[points.length - 1]['curve'][k];
                }
              }
            } catch (err) {
              _didIteratorError = true;
              _iteratorError = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }
              } finally {
                if (_didIteratorError) {
                  throw _iteratorError;
                }
              }
            }

            break;

          case 'C':
            points.push({
              curve: {
                type: 'cubic',
                x1: (relative ? prevPoint.x : 0) + commandParams.shift(),
                y1: (relative ? prevPoint.y : 0) + commandParams.shift(),
                x2: (relative ? prevPoint.x : 0) + commandParams.shift(),
                y2: (relative ? prevPoint.y : 0) + commandParams.shift()
              },
              x: (relative ? prevPoint.x : 0) + commandParams.shift(),
              y: (relative ? prevPoint.y : 0) + commandParams.shift()
            });

            break;

          case 'S':
            var sx2 = (relative ? prevPoint.x : 0) + commandParams.shift();
            var sy2 = (relative ? prevPoint.y : 0) + commandParams.shift();
            var sx = (relative ? prevPoint.x : 0) + commandParams.shift();
            var sy = (relative ? prevPoint.y : 0) + commandParams.shift();

            var diff = {};

            var sx1 = void 0;
            var sy1 = void 0;

            if (prevPoint.curve && prevPoint.curve.type === 'cubic') {
              diff.x = Math.abs(prevPoint.x - prevPoint.curve.x2);
              diff.y = Math.abs(prevPoint.y - prevPoint.curve.y2);
              sx1 = prevPoint.x < prevPoint.curve.x2 ? prevPoint.x - diff.x : prevPoint.x + diff.x;
              sy1 = prevPoint.y < prevPoint.curve.y2 ? prevPoint.y - diff.y : prevPoint.y + diff.y;
            } else {
              diff.x = Math.abs(sx - sx2);
              diff.y = Math.abs(sy - sy2);
              sx1 = prevPoint.x;
              sy1 = prevPoint.y;
            }

            points.push({ curve: { type: 'cubic', x1: sx1, y1: sy1, x2: sx2, y2: sy2 }, x: sx, y: sy });

            break;

          case 'Q':
            points.push({
              curve: {
                type: 'quadratic',
                x1: (relative ? prevPoint.x : 0) + commandParams.shift(),
                y1: (relative ? prevPoint.y : 0) + commandParams.shift()
              },
              x: (relative ? prevPoint.x : 0) + commandParams.shift(),
              y: (relative ? prevPoint.y : 0) + commandParams.shift()
            });

            break;

          case 'T':
            var tx = (relative ? prevPoint.x : 0) + commandParams.shift();
            var ty = (relative ? prevPoint.y : 0) + commandParams.shift();

            var tx1 = void 0;
            var ty1 = void 0;

            if (prevPoint.curve && prevPoint.curve.type === 'quadratic') {
              var _diff = {
                x: Math.abs(prevPoint.x - prevPoint.curve.x1),
                y: Math.abs(prevPoint.y - prevPoint.curve.y1)
              };

              tx1 = prevPoint.x < prevPoint.curve.x1 ? prevPoint.x - _diff.x : prevPoint.x + _diff.x;
              ty1 = prevPoint.y < prevPoint.curve.y1 ? prevPoint.y - _diff.y : prevPoint.y + _diff.y;
            } else {
              tx1 = prevPoint.x;
              ty1 = prevPoint.y;
            }

            points.push({ curve: { type: 'quadratic', x1: tx1, y1: ty1 }, x: tx, y: ty });

            break;
        }
      }
    } else {
      if (prevPoint.x !== moveTo.x || prevPoint.y !== moveTo.y) {
        points.push({ x: moveTo.x, y: moveTo.y });
      }
    }
  }

  return points;
};

var getPointsFromPolygon = function getPointsFromPolygon(_ref6) {
  var points = _ref6.points;

  return getPointsFromPoints({ closed: true, points: points });
};

var getPointsFromPolyline = function getPointsFromPolyline(_ref7) {
  var points = _ref7.points;

  return getPointsFromPoints({ closed: false, points: points });
};

var getPointsFromPoints = function getPointsFromPoints(_ref8) {
  var closed = _ref8.closed,
      points = _ref8.points;

  var numbers = points.split(/[\s,]+/).map(function (n) {
    return parseFloat(n);
  });

  var p = numbers.reduce(function (arr, point, i) {
    if (i % 2 === 0) {
      arr.push({ x: point });
    } else {
      arr[(i - 1) / 2].y = point;
    }

    return arr;
  }, []);

  if (closed) {
    p.push(_extends({}, p[0]));
  }

  p[0].moveTo = true;

  return p;
};

var getPointsFromRect = function getPointsFromRect(_ref9) {
  var height = _ref9.height,
      rx = _ref9.rx,
      ry = _ref9.ry,
      width = _ref9.width,
      x = _ref9.x,
      y = _ref9.y;

  if (rx || ry) {
    return getPointsFromRectWithCornerRadius({
      height: height,
      rx: rx || ry,
      ry: ry || rx,
      width: width,
      x: x,
      y: y
    });
  }

  return getPointsFromBasicRect({ height: height, width: width, x: x, y: y });
};

var getPointsFromBasicRect = function getPointsFromBasicRect(_ref10) {
  var height = _ref10.height,
      width = _ref10.width,
      x = _ref10.x,
      y = _ref10.y;

  return [{ x: x, y: y, moveTo: true }, { x: x + width, y: y }, { x: x + width, y: y + height }, { x: x, y: y + height }, { x: x, y: y }];
};

var getPointsFromRectWithCornerRadius = function getPointsFromRectWithCornerRadius(_ref11) {
  var height = _ref11.height,
      rx = _ref11.rx,
      ry = _ref11.ry,
      width = _ref11.width,
      x = _ref11.x,
      y = _ref11.y;

  var curve = { type: 'arc', rx: rx, ry: ry, sweepFlag: 1 };

  return [{ x: x + rx, y: y, moveTo: true }, { x: x + width - rx, y: y }, { x: x + width, y: y + ry, curve: curve }, { x: x + width, y: y + height - ry }, { x: x + width - rx, y: y + height, curve: curve }, { x: x + rx, y: y + height }, { x: x, y: y + height - ry, curve: curve }, { x: x, y: y + ry }, { x: x + rx, y: y, curve: curve }];
};

var getPointsFromG = function getPointsFromG(_ref12) {
  var shapes = _ref12.shapes;
  return shapes.map(function (s) {
    return toPoints(s);
  });
};

// Wilderness.js based shape converting tool
// @author @dalisoft

function morph (a, b, exportType) {
  if ( exportType === void 0 ) exportType = 'string';

  var string = exportType === 'string';
  var canvas = exportType === 'canvas';
  var group = a.type === 'g' && b.type === 'g';

  if (group) {
	var shape = {type: 'g', shapes: []};
	var l = Math.max(a.shapes.length, b.shapes.length);
	for (var i = 0; i < l; i++) {
		var sh1 = a.shapes[i];
		var sh2 = b.shapes[i];

		shape.shapes[i] = morph(sh1, sh2, exportType);
	}
	return shape;
  }

  var p1 = !!a ? toPoints(a) : false;
  var p2 = !!b ? toPoints(b) : false;

  if (p1 && !p2) {
	var ref = points_js.boundingBox(p1).center;
	var x = ref.x;
	var y = ref.y;
	p2 = [{x: x, y: y, moveTo: true}, {x: x, y: y}];
  } else if (p2 && !p1) {
	var ref$1 = points_js.boundingBox(p2).center;
	var x$1 = ref$1.x;
	var y$1 = ref$1.y;
	p1 = [{x: x$1, y: y$1, moveTo: true}, {x: x$1, y: y$1}];
  }

  var shapes = points_js.autoNormalise(p1, p2, {order:'auto'});
  var ref$2 = points_js.autoFix(shapes[0], shapes[1], {order:'auto'});
  var ap = ref$2[0];
  var bp = ref$2[1];
  var tpv, len = ap.length;

  if (a.moveIndex || b.moveIndex) {
	ap = points_js.moveIndex(ap, a.moveIndex || b.moveIndex);
  }
  if (a.reverse || b.reverse) {
	ap = points_js.reverse(ap);
  }
  
  if (canvas) {
	tpv = ap.map(function (ref) {
		var x = ref.x;
		var y = ref.y;
		var moveTo = ref.moveTo;
		var curve = ref.curve;

		if (curve) {
			curve = Object.assign({}, curve);
		}
		return {x: x, y: y, moveTo: moveTo, curve: curve}
	});
  }

  return function(t) {
    if (string) {
      var bufferString = '';
	  for (var i = 0; i < len; i++) {
		var ref = ap[i];
		var xf = ref.x;
		var yf = ref.y;
		var curve = ref.curve;
		var moveTo = ref.moveTo;
		var ref$1 = bp[i];
		var xt = ref$1.x;
		var yt = ref$1.y;
		var curve2 = ref$1.curve;
		var x = xf + (xt - xf) * t;
		var y = yf + (yt - yf) * t;
		if (moveTo) {
			bufferString += 'M' + x + ',' + y;
		} else if (curve && curve2) {
			var type = curve.type;
			var x1 = curve.x1;
			var y1 = curve.y1;
			var x2 = curve.x2;
			var y2 = curve.y2;
			var x3 = curve2.x1;
			var y3 = curve2.y1;
			var x4 = curve2.x2;
			var y4 = curve2.y2;
			if (type === 'cubic') {
				bufferString += 'C' + (x1 + (x3 - x1) * t) + ',' + (y1 + (y3 - y1) * t) + ',' + (x2 + (x4 - x2) * t) + ',' + (y2 + (y4 - y2) * t) + ',' + x + ',' + y;
			} else if (type === 'quadratic') {
				bufferString += 'Q' + (x1 + (x3 - x1) * t) + ',' + (y1 + (y3 - y1) * t) + ',' + x + ',' + y;
			}
		} else {
			bufferString += 'L' + x + ',' + y;
		}
	  }
	  return bufferString
    } else if (canvas) {
	  for (var i$1 = 0; i$1 < len; i$1++) {
		var ref$2 = ap[i$1];
		var xf$1 = ref$2.x;
		var yf$1 = ref$2.y;
		var curve$1 = ref$2.curve;
		var ref$3 = bp[i$1];
		var xt$1 = ref$3.x;
		var yt$1 = ref$3.y;
		var curve2$1 = ref$3.curve;

		var x$1 = xf$1 + (xt$1 - xf$1) * t;
		var y$1 = yf$1 + (yt$1 - yf$1) * t;

		tpv[i$1].x = x$1;
		tpv[i$1].y = y$1;

		if (curve$1 && curve2$1) {
			var ref$4 = curve$1;
			var type$1 = ref$4.type;
			var x1$1 = ref$4.x1;
			var y1$1 = ref$4.y1;
			var x2$1 = ref$4.x2;
			var y2$1 = ref$4.y2;
			var ref$5 = curve2$1;
			var x3$1 = ref$5.x1;
			var y3$1 = ref$5.y1;
			var x4$1 = ref$5.x2;
			var y4$1 = ref$5.y2;
			if (type$1 === 'cubic') {
				tpv[i$1].curve.x1 = (x1$1 + (x3$1 - x1$1) * t);
				tpv[i$1].curve.y1 = (y1$1 + (y3$1 - y1$1) * t);
				tpv[i$1].curve.x2 = (x2$1 + (x4$1 - x2$1) * t);
				tpv[i$1].curve.y2 = (y2$1 + (y4$1 - y2$1) * t);
			} else if (type$1 === 'quadratic') {
				tpv[i$1].curve.x1 = (x1$1 + (x3$1 - x1$1) * t);
				tpv[i$1].curve.y1 = (y1$1 + (y3$1 - y1$1) * t);
			}
		}
	  }
	  return tpv
    }
  }
}

morph.easing = {
	linear: function linear(k) { return k }
};

var index = Object.assign(morph, { draw: draw, frame: frame, middleware: middleware });

return index;

})));
//# sourceMappingURL=morph.js.map
